/* crackhead.c - crc32 bruteforcer
 * BSD-Licensed, 2010-2016, ed <irc.rizon.net>
 *
 * Compiling:
 * g++ -march=native -pipe -O3 -ffast-math -funroll-loops -funroll-all-loops -o crackhead crackhead.c
 *
 * Speed test:
 * for x in {1..5}; do for y in gcc gpp; do echo -n "$y $x   "; { time ./crackhead-$y 0x12341234 0xdeadbeef >/dev/null 2>/dev/null; } 2>&1 | sed -r 's/.*dev.null//'; done; done
 *
 * WARNING:
 * Usually SLOWER when compiling with gcc,
 * so use g++ whenever possible
 *
 * 2016/04/05:
 *   Compiles with gcc
 *   Changed license from GPLv2 to BSD
 *
 * 2010/02/10:
 *   Version 1.0
 */

#ifdef __cplusplus
#include <cstdio>
#include <cstdlib>
#include <sstream>
#else
#include <stdio.h>
#include <stdlib.h>
#endif
#define uint unsigned int


uint ctab[256];
uint csum(uint old, char item)
{
	return ((old >> 8) & 0x00ffffff)
		^ ctab[(old ^ item) & 0xff];
}

void cinit()
{
	int i, j;
	uint crc, poly;
	poly = 0xedb88320L;
	for (i = 0; i < 256; i++)
	{
		crc = i;
		for (j = 8; j > 0; j--)
		{
			if (crc & 1)
			{
				crc = (crc >> 1) ^poly;
			}
			else
			{
				crc >>= 1;
			}
		}
		ctab[i] = crc;
	}
}

uint htoi(char *str)
{
#ifdef __cplusplus
	std::stringstream ss;
	ss << std::hex << str;
	uint x; ss >> x;
	return static_cast<int>(x);
#else
	int ofs = 0;
	uint sum = 0;
	while (str[ofs] != 0)
	{
		char c = str[ofs++];
		sum *= 0x10;
		sum +=
			(c >= '0' && c <= '9') ? (c-'0') :
			(c >= 'a' && c <= 'f') ? (c-'a')+10 :
			(c >= 'A' && c <= 'F') ? (c-'A')+10 : 0;
	}
	return sum;
#endif
}

int main(int argc, char **argv)
{
	if (argc != 3)
	{
		printf("\n");
		printf("   crackhead by ed <tripflag%cgmail%ccom>\n",'@','.');
		printf("   Usage: %s OLDCRC NEWCRC >> file.mkv\n", argv[0]);
		printf("\n");
		printf("   Append output to original file\n");
		printf("   Data to append will be printed in hex to stderr\n");
		printf("   Data to append will be printed in raw to stdout\n");
		printf("\n");
		return 1;
	}
	//crca = 0xed82cd11;
	//crcb = 0x8587d865;

	uint ca,cb,cc;
	int ia,ib,ic,id;
	uint crca = htoi(argv[1]) ^ 0xffffffff;
	uint crcb = htoi(argv[2]) ^ 0xffffffff;

	char *cp = (char *)".oO@";
	fprintf(stderr,"[                                                                ]\r[");
	fflush(stderr);

	cinit();
	ca = crca;
	for (ia = -1; ia < 256; ia++)
	{
		// progressbar
		int ip = (ia + 1) % 4;
		if (ip==3)
		{
			fprintf(stderr,"%c",cp[ip]);
		}
		else
		{
			fprintf(stderr, "%c%c[D", cp[ip], 27);
		}
		fflush(stderr);

		// this looks ridiculous, but I can explain!
		// the overhead caused by keeping the checksum in an array
		// effectively doubles the processing time. so fuck that.

		if (ia >= 0)
		{
			ca = csum(crca, ia);
		}
		for (ib = -1; ib < 256; ib++)
		{
			if (ib < 0)
			{
				cb = ca;
			}
			else
			{
				cb = csum(ca, ib);
			}
			for (ic = -1; ic < 256; ic++)
			{
				if (ic < 0)
				{
					cc = cb;
				}
				else
				{
					cc = csum(cb, ic);
				}
				for (id = 0; id < 256; id++)
				{
					if (csum(cc, id) == crcb)
					{
						if (ia >= 0) printf("%c", (unsigned char)ia);
						if (ib >= 0) printf("%c", (unsigned char)ib);
						if (ic >= 0) printf("%c", (unsigned char)ic);
						printf("%c", (unsigned char)id);

						fprintf(stderr, "\n");
						if (ia >= 0) fprintf(stderr, "%02x", ia);
						if (ib >= 0) fprintf(stderr, "%02x", ib);
						if (ic >= 0) fprintf(stderr, "%02x", ic);
						fprintf(stderr, "%02x\n", id);
						return 0;
					}
				}
			}
		}
	}
	fprintf(stderr, "\nwait what\n");
	return 1;
}
